/*

	CTC.C

	The Counter/Timer Chip

   by Oliver Kraus

	most parts based on:
	Part of the PC Timing FAQ / Application notes
	By K. Heidenstrom (kheidens@actrix.gen.nz)

	tested with WATCOM C 10.0, MSC 7.0, GCC 2.6.2


	int ctc_detect_mode(void)

	Set up timing functions. This function is automaticly called,
	by the first call to ctc_time() or ctc_pause().
	If ctc_detect_mode() returns CTC_MODE_ILLEGAL resolution
	of ctc_time() and ctc_pause() is only 18.2 Hz.

																					
	unsigned long ctc_time(void)

	Get time since midnight in CTC_TICKS_PER_SEC
	Example: 
		printf("Seconds since midnight: %ld\n",  ctc_time()/CTC_TICKS_PER_SEC);


	void ctc_pause(unsigned long delay)

	Delay programm execution by specified value.
	Example: Delay 2 seconds
		ctc_pause(2UL*CTC_TICKS_PER_SEC);
	Example: Delay 150ms
		ctc_pause((CTC_TICKS_PER_SEC*15UL)/100UL);

*/

#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include "ctc.h"
#ifdef __GNUC__
#include <sys\farptr.h>
#endif 


#define TESTVALUE 0x055AA   /* Value to use as reload test value */
#define BACKWARDS 0x0AA55   /* Backwards TESTVALUE */
#define EXP_CSTAT 0x30     /* Expected counter status */

#define CTC_EMUL  0  /* CTC is faulty or emulated by OS */
#define CTC_8253  1  /* CTC is an 8253 */
#define CTC_8254  2  /* CTC is an 8254 */

typedef struct
{
	unsigned int count;
	unsigned char status;
} readback_data;

#ifdef __GNUC__
#define inp(port) ((unsigned)inportb(port))
#define outp(port,val) outportb(port, (unsigned char)(val))
#define _enable() enable()
#define _disable() disable()
#endif

int ctc_detected_mode = CTC_MODE_NONE;

static unsigned int read_channel_raw(unsigned int channum)
{
	unsigned int cv;
	outp(0x43, channum << 6);
	cv = inp(0x40 + channum);
	cv += inp(0x40 + channum) << 8;
	return cv;
}

/* Simple short delay - just wait for at least one CTC clock to occur */
static void wait_ctc_clock(void)
{
	unsigned int ch0count;
	ch0count = read_channel_raw(0);
	while (read_channel_raw(0) == ch0count)
		;
	return;
}

static void readback_channel(unsigned int channum, readback_data * rbdp)
{
   outp(0x43, 0xC0 + (2 << channum));        /* Latch count, status */
	rbdp->status = (unsigned char)inp(0x40 + channum); /* Get status */
   rbdp->count = inp(0x40 + channum);        /* Get count lobyte */
   rbdp->count += inp(0x40 + channum) << 8;  /* Get count hibyte */
}

/* This function determines the CTC type. */
unsigned int ctc_detect_type(void)
{
	unsigned int ctctype;               /* Value to be returned */
	unsigned int port61;                /* Port 61h value */
	readback_data ch2rbd, rbd1, rbd2;   /* Read-back storage */
	unsigned int backwards, forwards;   /* Latched count values */

	ctctype = CTC_EMUL;                 /* Assume faulty or unknown CTC type */

	_disable();

	port61 = inp(0x61);                 /* Get Port B value */
	outp(0x61, port61 & 0xFC);          /* Turn off timer 2 gate and speaker */

/* Try read-back on channel two, only useful if CTC turns out to be an 8254 */

	readback_channel(2, &ch2rbd);
	readback_channel(2, &ch2rbd);       /* Attempt to read-back channel two */

   outp(0x43, 0xB0);                   /* Cha. 2, two bytes, mode 0, binary */
	outp(0x42, TESTVALUE & 0xFF);       /* Lobyte of reload value */
	outp(0x42, TESTVALUE >> 8);         /* Hibyte of reload value */

	wait_ctc_clock();
	wait_ctc_clock(); /* Wait for a couple of CTC clock pulses */

/* Just read the raw value a couple of times, make sure it's stable */

	if ((read_channel_raw(2) != TESTVALUE) ||
		(read_channel_raw(2) != TESTVALUE))
		goto got_type; /* Structured programming?  Never heard of it */

/* Try a read-back - on an 8253, this will reverse the lobyte/hibyte flag */

	readback_channel(2, &rbd1);

/* Read the count - on an 8253 this will be TESTVALUE backwards, on an 8254
	it will be TESTVALUE */

	backwards = read_channel_raw(2);

/* Try another read-back, into rbd2 this time */

	readback_channel(2, &rbd2);

/* Now latch and read the count again */

	forwards = read_channel_raw(2);

/* Now, try to figure out what it is! */

	if ((rbd1.status != EXP_CSTAT) && (rbd2.status != EXP_CSTAT) &&
		(backwards == BACKWARDS) && (forwards == TESTVALUE))
			ctctype = CTC_8253;
	if ((rbd1.status == EXP_CSTAT) && (rbd2.status == EXP_CSTAT) &&
		(backwards == TESTVALUE) && (forwards == TESTVALUE))
			ctctype = CTC_8254;
got_type:

/* Now we know what it is.  If it's an 8254, we can restore channel 2 to its
	previous mode, although we cannot restore the original divisor, because
	we can't tell what it was.  If it's not an 8254, we can't fix anything */

	if (ctctype == CTC_8254)
	{
		outp(0x43, 0x80 + (ch2rbd.status & 0x3F));
		outp(0x42, 0);
		outp(0x42, 0);
	}

	outp(0x61, port61); /* Restore speaker and timer 2 control bits */

	_enable();
	return ctctype;
}



unsigned short ctc_read_mode3(int do_disable)
{
	unsigned short st1, st2;               /* Status read-back values */
	unsigned short cv;                     /* Count value */
	if (do_disable)
		_disable();                         /* No interrupts please */
	outp(0x43, 0xE2);                      /* Latch and read back status byte */
	st1 = (unsigned short)inp(0x40);       /* Read status byte */
	outp(0x43, 0x00);                      /* Latch count for timer 0 */
	cv = (unsigned short)inp(0x40);        /* Lobyte of count */
	cv += (unsigned short)(inp(0x40) << 8);/* Hibyte of count */
	cv = (unsigned short)((0 - cv) >> 1);  /* Convert to up-count, 0-32767 */
	outp(0x43, 0xE2);                      /* Latch and read back status byte */
	st2 = (unsigned short)inp(0x40);       /* Read status byte */
	if (do_disable)
		_enable();                          /* Interrupts back on */
	if ((st1 ^ st2) & 0x80)                /* If output pin changed state... */
		if (cv < 0x4000)                    /* If reload just occurred... */
			st1 ^= 0x80;                     /* Use newer output pin status */
	if ((st1 & 0x80) == 0)                 /* If on second countdown... */
		cv |= 0x8000;                       /* Set b15 */
	return cv;                             /* Return as up-counter */
}


void ctc_force_mode2(void)
{
	_disable();
	outp(0x43, 0x34);      /* Channel 0, mode 2 */
	outp(0x40, 0x00);      /* Loword of divisor */
	outp(0x40, 0x00);      /* Hiword of divisor */
	_enable();
}

unsigned short ctc_read_mode2(int do_disable)
{
	unsigned short val;
	if (do_disable)
		_disable();
	outp(0x43, 0);             /* Latch Command */
	val = (unsigned short)inp(0x40);
	val += (unsigned short)(inp(0x40) << 8);
	if (do_disable)
		_enable();
	return (unsigned short)((unsigned short)0-val);
}


unsigned long bios_read_ticks(void)
{
	unsigned long val;
	_disable();
#ifdef __WATCOMC__
#ifdef __386__
	val = *((unsigned long *)0x046CL);     /* assuming DOS4GW linear memory */
#else
	val = *((unsigned long far *)0x0040006CL);   /* WATCOM 16 Bit Compiler */
#endif
#else
#ifdef __GNUC__
	val = _farpeekl(_go32_conventional_mem_selector(), 0x046CL);
#else
	val = *((unsigned long far *)0x0040006CL);   /* any other 16 Bit Compiler */
#endif
#endif
	_enable();
	return val;
}

int ctc_detect_mode(void)
{
	unsigned long ticks;
	unsigned int mode;
	int ov = 0;

	ctc_detected_mode = CTC_MODE_NONE;
	mode = ctc_detect_type();
	if ( mode == CTC_EMUL )
	{
		ctc_detected_mode = CTC_MODE_ILLEGAL;
		return ctc_detected_mode;
	}
   else if ( mode == CTC_8254 )
	{
		while(ctc_read_mode2(1) > 0x08000)
			;
		ticks = bios_read_ticks();
		while(ctc_read_mode2(1) <= 0x08000)
			;
		while(ctc_read_mode2(1) > 0x08000)
			;
		if ( ticks != bios_read_ticks() )
			ov++;
		ticks = bios_read_ticks();
		while(ctc_read_mode2(1) <= 0x08000)
			;
		while(ctc_read_mode2(1) > 0x08000)
			;
		if ( ticks != bios_read_ticks() )
			ov++;
		if ( ov == 1 )
			ctc_detected_mode = CTC_MODE_3;
		else if ( ov == 2 )
			ctc_detected_mode = CTC_MODE_2;
	}
   else 
	{
		ctc_force_mode2();
		ctc_detected_mode = CTC_MODE_2;
	}
	return ctc_detected_mode;
}


/*
	read timer latch 0-65535
*/
unsigned short ctc_read(void)
{
	if ( ctc_detected_mode == CTC_MODE_NONE )
		ctc_detect_mode();
	if ( ctc_detected_mode == CTC_MODE_3 )
		return ctc_read_mode3(1);
	if ( ctc_detected_mode == CTC_MODE_2 )
		return ctc_read_mode2(1);
	return 0;
}

unsigned long ctc_time(void)
{
	unsigned long t1, t2;
	unsigned short cnt;
	if ( ctc_detected_mode == CTC_MODE_NONE )
		ctc_detect_mode();
	t1 = bios_read_ticks();
	cnt = ctc_read();
	t2 = bios_read_ticks();
	if ( t1 != t2 && cnt < 0x08000)
		t1 = t2;
	t1 = (t1<<10) | (cnt>>6);
	return t1;
}



void ctc_pause(unsigned long delay)
{
	unsigned long t1, t2;
	t1 = ctc_time();
	for(;;)
	{
		t2 = ctc_time();
		if ( t1 <= t2 )
		{
			if ( t2 > delay + t1 )
				break;
		}
		else
		{
			if ( CTC_TICKS_MAX + t2 > delay + t1 )
				break;
		}
	}
}

/* #define test_code */
#ifdef test_code

void main(void)
{

	unsigned int type;
	type = ctc_detect_type();
	switch(type)
	{
		case CTC_8253:
			puts("Timer: 8253");
			break;
		case CTC_8254:
			puts("Timer: 8254");
			break;
		default:
			puts("Timer: unknown");
			break;
	}

	ctc_detect_mode();
	if ( ctc_detected_mode == CTC_MODE_NONE )
		puts("illegal timer mode");
	if ( ctc_detected_mode == CTC_MODE_3 )
		puts("timer mode 3");
	if ( ctc_detected_mode == CTC_MODE_2 )
		puts("timer mode 2");

	getch();

	for(;;)
		printf("%16ld",
			(unsigned long)ctc_time());
}

#endif
